/*
	Sudoku generator

	Windows interface

	(C) Copyright 2005 by Mogens Kjaer, mk@lemo.dk

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdarg.h>
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>

#include "sudoku.h"
#include "sudokures.h"

static HWND mainhwnd=NULL, printerhwnd=NULL, settingshwnd=NULL;
static HINSTANCE mainhInstance;

int nfilled=28;
int newnfilled, newsymmetrical, newlevel, newnx, newny;
int bestndigits=0;
int ndigits;
int symmetrical=1;
int level=LEVEL_EASY;
char symbollist[]="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";

int statistics[MAXNXNY*MAXNXNY][NLEVELS];
static time_t start_time, stop_time;

#define THINWIDTH 1
#define THICKWIDTH 3
#define CELLSIZE 50

int displayboard[MAXNXNY][MAXNXNY];

int setboard[MAXNXNY][MAXNXNY];
int setboardsave[MAXNXNY][MAXNXNY];
int beforeboard[MAXNXNY][MAXNXNY];
int board[MAXNXNY][MAXNXNY];
int nx=3;
int ny=3;
int nxny=9;
int nsol=0;
int ntest=0;
int running=0;
int solved=0;
int solution_unique=0;

static HFONT font, oldfont;
static int default_textsize=CELLSIZE;
static int font_width, font_height;
static HANDLE printer_hDevMode, printer_hDevNames;
static RECT printer_lastMargin;
static int printer_marginunits;
static int print_explain=1;

int nlines=0;
char *loglines=NULL;
int logsize=0;

void loginit()
{
  if(loglines!=NULL) free(loglines);
  loglines=NULL;
  logsize=0;
  nlines=0;
}

void logend()
{
}

void logprintf(const char *template, ...)
{
  va_list ap;
  char tmpbuf[1024];
  int len,i;

  va_start(ap, template);
  vsprintf(tmpbuf, template, ap);
  va_end(ap);
  len=strlen(tmpbuf);
  for(i=0;i<len;i++)
  {
    if(tmpbuf[i]=='\n') nlines++;
  }
  
  if(loglines==NULL)
  {
    loglines=malloc(len+1);
    loglines[0] = '\0';
  }
  else
  {
    loglines=realloc(loglines, strlen(loglines)+len+1);
  }
  strcat(loglines, tmpbuf);
}

static void setmenutext(int item, char *txt)
{
  MENUITEMINFO menuiteminfo;

  menuiteminfo.cbSize = sizeof(MENUITEMINFO);
  menuiteminfo.fMask = MIIM_STRING;
  menuiteminfo.fType = MFT_STRING;
  menuiteminfo.dwTypeData = txt;
  menuiteminfo.cch = strlen(txt);

  SetMenuItemInfo(GetMenu(mainhwnd), item, FALSE, &menuiteminfo);

  DrawMenuBar(mainhwnd);

}

setcreatemenutext(char *txt)
{
  setmenutext(IDM_MAIN_FILE_CREATE, txt);
}

setsolvemenutext(char *txt)
{
  setmenutext(IDM_MAIN_FILE_SOLVE, txt);
}

settitle(char *txt)
{
  SetWindowText(mainhwnd, txt);
}

static void generate_sudoku()
{
  bestndigits=0;
  running=1;
  ntest=0;
}

static printer_print(HWND hWnd)
{
  PRINTDLG pd;
  HDC hdc;
  HPEN hpen, hpenold;
  DOCINFO di;
  SIZE tsize;
  BOOL bSuccess = TRUE;
  BOOL bUserAbort = FALSE;
  int iPhysWidth, iPhysHeight, iPhysOffsetX, iPhysOffsetY;
  int mleft, mtop, mright, mbottom;
  int iLeftAdjust, iTopAdjust;
  int fontsize, explainfontsize;
  int printer_font_width, printer_font_height;
  int lpp1, lpp2;
  TEXTMETRIC tm;
  HFONT printer_font, explain_printer_font;
  HANDLE hold;
  int iColCopy, iNonColCopy, iPage, iLine, iLinePage, iCol, iChar;
  HBRUSH hbr, hbrold;
  LOGBRUSH logbrush;
  POINT oldorigin;
  int oldmode;
  int printer_thinwidth, printer_thickwidth, printer_cellsize;

#ifdef UAKK
  FILE *debug_fh;
#endif

  int width, x, y, row, col, size, yspacing, i;
  int npages, height1, height2;
  char txt[2];
  char *ploglines;

  logbrush.lbStyle = BS_SOLID;
  logbrush.lbColor = RGB(0,0,0);
  logbrush.lbHatch = 0;

  /* Hm, we don't know the number of pages until we have selected the printer... */

  memset(&pd, 0, sizeof(pd));
  pd.lStructSize = sizeof(pd);
  pd.hwndOwner = hWnd;
  pd.hDevMode = printer_hDevMode;
  pd.hDevNames = printer_hDevNames;
  pd.nMinPage = 1;
  pd.nMaxPage = 1;
  pd.nFromPage = 1;
  pd.nToPage = 1;
  pd.Flags = PD_RETURNDC;
  if(PrintDlg(&pd))
  {
    printer_hDevMode = pd.hDevMode;
    printer_hDevNames = pd.hDevNames;
    hdc = pd.hDC;
    mleft = MulDiv(printer_lastMargin.left, GetDeviceCaps(hdc, LOGPIXELSX), 1000);
    mtop = MulDiv(printer_lastMargin.top, GetDeviceCaps(hdc, LOGPIXELSY), 1000);
    mright = MulDiv(printer_lastMargin.right, GetDeviceCaps(hdc, LOGPIXELSX), 1000);
    mbottom = MulDiv(printer_lastMargin.bottom, GetDeviceCaps(hdc, LOGPIXELSY), 1000);
    iPhysOffsetX = GetDeviceCaps(hdc, PHYSICALOFFSETX);
    iPhysOffsetY = GetDeviceCaps(hdc, PHYSICALOFFSETY);
    iPhysWidth = GetDeviceCaps(hdc, PHYSICALWIDTH);
    iPhysHeight = GetDeviceCaps(hdc, PHYSICALHEIGHT);

    if(GetDeviceCaps(hdc, VERTRES)>GetDeviceCaps(hdc, HORZRES))
    {
      size = GetDeviceCaps(hdc, HORZRES);
    }
    else
    {
      size = GetDeviceCaps(hdc, VERTRES);
    }
    size = size*4/5;
    if(solved && print_explain && solution_unique==1)
    {
      /* Print out explanation: */
      unsolve_sudoku();
      loginit();
      solve_sudoku(1);
      logend();
      size = size/3;
      iTopAdjust = mtop;
      iLeftAdjust = (GetDeviceCaps(hdc, HORZRES)-size)/2;
      yspacing = 15*GetDeviceCaps(hdc, LOGPIXELSY)/72;
      height1 = GetDeviceCaps(hdc, VERTRES)-size-mtop-mbottom;
      height2 = GetDeviceCaps(hdc, VERTRES)-mtop-mbottom;
      /* Lines per page, first page: */
      lpp1 = height1/yspacing;
      /* Following pages: */
      lpp2 = height2/yspacing;
      if(nlines > lpp1)
      {
	npages = 1 + (nlines-lpp1+lpp2-1)/lpp2;
      }
      else
      {
        npages=1;
      }
      explainfontsize = yspacing*4/5;
      explain_printer_font = CreateFont(-explainfontsize,0,0,0,400,0,0,0,ANSI_CHARSET,
              OUT_TT_PRECIS,CLIP_DEFAULT_PRECIS,
	      PROOF_QUALITY,FIXED_PITCH|FF_MODERN,"Courier New");
    }
    else
    {
      iTopAdjust = (GetDeviceCaps(hdc, VERTRES)-size)/2;
      iLeftAdjust = (GetDeviceCaps(hdc, HORZRES)-size)/2;
      npages=1;
      nlines=1;
    }
    printer_cellsize = size/nxny;
    fontsize = -printer_cellsize*10/11;
    printer_thinwidth = printer_cellsize/40;
    printer_thickwidth = printer_thinwidth*3;
#ifdef UAKK
    debug_fh = fopen("sudoku.debug", "w");
    fprintf(debug_fh, "iPhysOffsetX: %d\n", iPhysOffsetX);
    fprintf(debug_fh, "iPhysOffsetY: %d\n", iPhysOffsetY);
    fprintf(debug_fh, "iPhysWidth: %d\n", iPhysWidth);
    fprintf(debug_fh, "iPhysHeight: %d\n", iPhysHeight);
    fprintf(debug_fh, "mleft: %d\n", mleft);
    fprintf(debug_fh, "mright: %d\n", mright);
    fprintf(debug_fh, "mtop: %d\n", mtop);
    fprintf(debug_fh, "mbottom: %d\n", mbottom);
    fprintf(debug_fh, "iLeftAdjust: %d\n", iLeftAdjust);
    fprintf(debug_fh, "iTopAdjust: %d\n", iTopAdjust);
    fprintf(debug_fh, "LOGPIXELSX: %d\n", GetDeviceCaps(hdc, LOGPIXELSX));
    fprintf(debug_fh, "LOGPIXELSY: %d\n", GetDeviceCaps(hdc, LOGPIXELSY));
    fprintf(debug_fh, "HORZRES: %d\n", GetDeviceCaps(hdc, HORZRES));
    fprintf(debug_fh, "VERTRES: %d\n", GetDeviceCaps(hdc, VERTRES));
    fprintf(debug_fh, "size: %d\n", size);
    fprintf(debug_fh, "fontsize: %d\n", fontsize);
    fprintf(debug_fh, "explainfontsize: %d\n", explainfontsize);
    fprintf(debug_fh, "nlines: %d\n", nlines);
    fprintf(debug_fh, "npages: %d\n", npages);
    fprintf(debug_fh, "yspacing: %d\n", yspacing);
    fprintf(debug_fh, "lpp1: %d\n", lpp1);
    fprintf(debug_fh, "lpp2: %d\n", lpp2);
    fprintf(debug_fh, "height1: %d\n", height1);
    fprintf(debug_fh, "height2: %d\n", height2);
    fclose(debug_fh);
#endif
    SetMapMode(hdc, MM_TEXT);
    printer_font = CreateFont(-fontsize,0,0,0,400,0,0,0,ANSI_CHARSET,OUT_TT_PRECIS,CLIP_DEFAULT_PRECIS,
		      PROOF_QUALITY,FIXED_PITCH|FF_MODERN,"Courier New");
    
    hold=SelectObject(hdc, printer_font);
    GetTextMetrics(hdc, &tm);
    GetTextExtentPoint32(hdc, "W", 1, &tsize);
    SelectObject(hdc, hold);
    printer_font_width = tsize.cx;
    printer_font_height = tsize.cy;

    memset(&di, 0, sizeof(di));
    di.cbSize = sizeof(di);
    di.lpszDocName = "Sudoku Print";
    oldmode=SetBkMode(hdc, TRANSPARENT);
    hbr = GetStockObject(NULL_BRUSH);
    hbrold = SelectObject(hdc, hbr);


    if(StartDoc(hdc, &di)>0)
    {
      for(iColCopy=0; iColCopy<((pd.Flags&PD_COLLATE)?pd.nCopies:1); iColCopy++)
      {
	iPage = 0;
	iLine = 0;
	ploglines = loglines;
        while(iLine < nlines)
	{
	  iPage++;
	  if((!(pd.Flags&PD_PAGENUMS)) ||
	     (iPage>=pd.nFromPage && iPage<=pd.nToPage))
	  {
	    for(iNonColCopy = 0; iNonColCopy < ((pd.Flags&PD_COLLATE)?1:pd.nCopies); iNonColCopy++)
	    {
	      if(StartPage(hdc)<0)
	      {
		bSuccess = FALSE;
		break;
	      }
	      if(iPage==1)
	      {
		SetViewportOrgEx(hdc, iLeftAdjust, iTopAdjust, &oldorigin);
		y=0;
		for(row=0;row<nxny;row++)
		{
		  if(row>0)
		  {
		    width = (row%ny==0)?printer_thickwidth:printer_thinwidth;
		    hpen = ExtCreatePen(PS_GEOMETRIC | PS_SOLID | PS_ENDCAP_SQUARE | PS_JOIN_BEVEL,
					width, &logbrush, 0, NULL);
		    hpenold = SelectObject(hdc, hpen);
		    MoveToEx(hdc, y, 0, NULL);
		    LineTo(hdc, y, size);
		    SelectObject(hdc, hpenold);
		    DeleteObject(hpen);
		    width = (row%nx==0)?printer_thickwidth:printer_thinwidth;
		    hpen = ExtCreatePen(PS_GEOMETRIC | PS_SOLID | PS_ENDCAP_SQUARE | PS_JOIN_BEVEL,
					width, &logbrush, 0, NULL);
		    hpenold = SelectObject(hdc, hpen);
		    MoveToEx(hdc, 0, y, NULL);
		    LineTo(hdc, size, y);
		    SelectObject(hdc, hpenold);
		    DeleteObject(hpen);
		  }
		  x=0;
		  for(col=0;col<nxny;col++)
		  {
		    txt[1]=0;
		    txt[0]=displayboard[row][col]?symbollist[abs(displayboard[row][col])]:' ';
		    oldfont=SelectObject(hdc, printer_font);
		    SetTextColor(hdc, displayboard[row][col]<0?RGB(255,0,0):RGB(0,0,0));
		    TextOut(hdc, x+(printer_cellsize-printer_font_width)/2, y, txt, 1);
		    SelectObject(hdc, oldfont);
		    x+=printer_cellsize;
		  }
		  y+=printer_cellsize;
		}
		hpen = ExtCreatePen(PS_GEOMETRIC | PS_SOLID | PS_ENDCAP_SQUARE | PS_JOIN_BEVEL,
				    printer_thickwidth, &logbrush, 0, NULL);
		hpenold = SelectObject(hdc, hpen);
		BeginPath(hdc);
		MoveToEx(hdc, 0, 0, NULL);
		LineTo(hdc, 0, size);
		LineTo(hdc, size, size);
		LineTo(hdc, size, 0);
		CloseFigure(hdc);
		EndPath(hdc);
		StrokePath(hdc);
		SelectObject(hdc, hpenold);
		DeleteObject(hpen);
	      }
	      if(solved && print_explain && solution_unique==1)
	      {
	        if(iPage==1)
		{
		  iTopAdjust = size+mtop;
		  iLeftAdjust = mleft;
		}
		else
		{
		  iTopAdjust = mtop;
		  iLeftAdjust = mleft;
		}
		SetViewportOrgEx(hdc, iLeftAdjust, iTopAdjust, NULL);
		y=0;
		iLinePage=0;
		for(;;)
		{
		  for(i=0;;i++)
		  {
		    if(ploglines[i]=='\n' || ploglines[i]=='\0') break;
		  }
		  oldfont=SelectObject(hdc, explain_printer_font);
		  SetTextColor(hdc, RGB(0,0,0));
		  TextOut(hdc, 0, y, ploglines, i);
		  SelectObject(hdc, oldfont);
		  if(ploglines[i]=='\0') break;
		  ploglines += i+1;
		  iLine++;
		  iLinePage++;
		  if(iPage==1)
		  {
		    if(iLinePage>=lpp1) break;
		  }
		  else
		  {
		    if(iLinePage>=lpp2) break;
		  }
		  y += yspacing;
		}
	      }
	      else
	      {
	        iLine=2;
	      }
	    }
	    if(EndPage(hdc)<0)
	    {
	      bSuccess = FALSE;
	      break;
	    }
	  }
	  else
	  {
	    /* skip forward in ploglines */
	    iLinePage=0;
	    for(;;)
	    {
	      for(i=0;;i++)
	      {
		if(ploglines[i]=='\n' || ploglines[i]=='\0') break;
	      }
	      if(ploglines[i]=='\0') break;
	      ploglines += i+1;
	      iLine++;
	      iLinePage++;
	      if(iPage==1)
	      {
		if(iLinePage>=lpp1) break;
	      }
	      else
	      {
		if(iLinePage>=lpp2) break;
	      }
	    }
	  }
	  if(!bSuccess || bUserAbort) break;
	}
	if(!bSuccess || bUserAbort) break;
      }
      EndDoc(hdc);
    }

    SelectObject(hdc, hbrold);
    DeleteObject(hbr);
    SetBkMode(hdc, oldmode);
    DeleteObject(printer_font);
    DeleteObject(explain_printer_font);
    DeleteDC(hdc);
  }
}

BOOL CALLBACK PrinterDlgProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  int i;
  LPDEVNAMES lpdevnames;
  PAGESETUPDLG pdlg;
  switch(message)
  {
  case WM_INITDIALOG:
    lpdevnames = (LPDEVNAMES) GlobalLock(printer_hDevNames);
    SetDlgItemText(hWnd, IDM_PRINTER_NAME, ((char *) lpdevnames)+lpdevnames->wDeviceOffset);
    GlobalUnlock(lpdevnames);
    SendMessage(GetDlgItem(hWnd, IDM_PRINTER_EXPLAIN), BM_SETCHECK, print_explain?BST_CHECKED:BST_UNCHECKED, 0);
    break;
  case WM_COMMAND:
    switch(LOWORD(wParam))
    {
    case IDM_PRINTER_SELECT:
      memset(&pdlg, 0, sizeof(pdlg));
      pdlg.lStructSize = sizeof(pdlg);
      pdlg.hwndOwner = mainhwnd;
      pdlg.hDevMode = printer_hDevMode;
      pdlg.hDevNames = printer_hDevNames;
      pdlg.rtMargin.top = printer_lastMargin.top;
      pdlg.rtMargin.bottom = printer_lastMargin.bottom;
      pdlg.rtMargin.left = printer_lastMargin.left;
      pdlg.rtMargin.right = printer_lastMargin.right;
      pdlg.Flags = PSD_MARGINS|printer_marginunits;
      PageSetupDlg(&pdlg);
      printer_hDevMode = pdlg.hDevMode;
      printer_hDevNames = pdlg.hDevNames;
      printer_lastMargin.top = pdlg.rtMargin.top;
      printer_lastMargin.bottom = pdlg.rtMargin.bottom;
      printer_lastMargin.left = pdlg.rtMargin.left;
      printer_lastMargin.right = pdlg.rtMargin.right;
      printer_marginunits = pdlg.Flags&(PSD_INTHOUSANDTHSOFINCHES|PSD_INHUNDREDTHSOFMILLIMETERS);
      lpdevnames = (LPDEVNAMES) GlobalLock(printer_hDevNames);
      SetDlgItemText(hWnd, IDM_PRINTER_NAME, ((char *) lpdevnames)+lpdevnames->wDeviceOffset);
      GlobalUnlock(lpdevnames);
      break;
    case IDM_PRINTER_EXPLAIN:
      print_explain = SendMessage(GetDlgItem(hWnd, IDM_PRINTER_EXPLAIN), BM_GETCHECK, 0, 0) == BST_CHECKED;
      break;
    case IDM_PRINTER_PRINT:
      EndDialog(hWnd, IDM_PRINTER_PRINT);
      return TRUE;
    case IDM_PRINTER_CANCEL:
      EndDialog(hWnd, IDM_PRINTER_CANCEL);
      return TRUE;
    default:
      return TRUE;
    }
  default:
    return FALSE;
  }
  return FALSE;
}

static void set_defaultnfilled(HWND hWnd)
{
  int newnxny,defaultnfilled;
  newnxny=newnx*newny;
  defaultnfilled=0;
  if(newnxny==4) defaultnfilled=8;
  else if(newnxny==6) defaultnfilled=12;
  else if(newnxny==8) defaultnfilled=20;
  else if(newnxny==9) defaultnfilled=28;
  else if(newnxny==10) defaultnfilled=36;
  else if(newnxny==12) defaultnfilled=56;
  else if(newnxny==15) defaultnfilled=100;
  else if(newnxny==16) defaultnfilled=112;
  else if(newnxny==20) defaultnfilled=200;
  if(defaultnfilled>0) SetDlgItemInt(hWnd, IDM_SETTINGS_FILLED, defaultnfilled, FALSE);
}

BOOL CALLBACK SettingsDlgProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  char tmp_label[80];
  int i;
  switch(message)
  {
  case WM_INITDIALOG:
    newnfilled = nfilled;
    newsymmetrical = symmetrical;
    newlevel = level;
    newnx = nx;
    newny = ny;
    SetDlgItemInt(hWnd, IDM_SETTINGS_FILLED, newnfilled, FALSE);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_NOTSYMMETRICAL), BM_SETCHECK, (newsymmetrical==0)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_XYSYMMETRICAL), BM_SETCHECK, (newsymmetrical==1)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_ROTSYMMETRICAL), BM_SETCHECK, (newsymmetrical==2)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_VERYEASY), BM_SETCHECK, (newlevel==LEVEL_VERYEASY)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_EASY), BM_SETCHECK, (newlevel==LEVEL_EASY)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_MEDIUM), BM_SETCHECK, (newlevel==LEVEL_MEDIUM)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_HARD), BM_SETCHECK, (newlevel==LEVEL_HARD)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_DIFFICULT), BM_SETCHECK, (newlevel==LEVEL_DIFFICULT)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_NX_2), BM_SETCHECK, (newnx==2)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_NX_3), BM_SETCHECK, (newnx==3)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_NX_4), BM_SETCHECK, (newnx==4)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_NY_2), BM_SETCHECK, (newny==2)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_NY_3), BM_SETCHECK, (newny==3)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_NY_4), BM_SETCHECK, (newny==4)?BST_CHECKED:BST_UNCHECKED, 0);
    SendMessage(GetDlgItem(hWnd, IDM_SETTINGS_NY_5), BM_SETCHECK, (newny==5)?BST_CHECKED:BST_UNCHECKED, 0);
    break;
  case WM_COMMAND:
    switch(LOWORD(wParam))
    {
    case IDM_SETTINGS_FILLED:
      newnfilled = GetDlgItemInt(hWnd, IDM_SETTINGS_FILLED, NULL, FALSE);
      break;
    case IDM_SETTINGS_NOTSYMMETRICAL:
      newsymmetrical = 0;
      break;
    case IDM_SETTINGS_XYSYMMETRICAL:
      newsymmetrical = 1;
      break;
    case IDM_SETTINGS_ROTSYMMETRICAL:
      newsymmetrical = 2;
      break;
    case IDM_SETTINGS_VERYEASY:
      newlevel = LEVEL_VERYEASY;
      break;
    case IDM_SETTINGS_EASY:
      newlevel = LEVEL_EASY;
      break;
    case IDM_SETTINGS_MEDIUM:
      newlevel = LEVEL_MEDIUM;
      break;
    case IDM_SETTINGS_HARD:
      newlevel = LEVEL_HARD;
      break;
    case IDM_SETTINGS_DIFFICULT:
      newlevel = LEVEL_DIFFICULT;
      break;
    case IDM_SETTINGS_NX_2:
      newnx=2;
      set_defaultnfilled(hWnd);
      break;
    case IDM_SETTINGS_NX_3:
      newnx=3;
      set_defaultnfilled(hWnd);
      break;
    case IDM_SETTINGS_NX_4:
      newnx=4;
      set_defaultnfilled(hWnd);
      break;
    case IDM_SETTINGS_NY_2:
      newny=2;
      set_defaultnfilled(hWnd);
      break;
    case IDM_SETTINGS_NY_3:
      newny=3;
      set_defaultnfilled(hWnd);
      break;
    case IDM_SETTINGS_NY_4:
      newny=4;
      set_defaultnfilled(hWnd);
      break;
    case IDM_SETTINGS_NY_5:
      newny=5;
      set_defaultnfilled(hWnd);
      break;
    case IDM_SETTINGS_OK:
      EndDialog(hWnd, IDM_SETTINGS_OK);
      return TRUE;
    case IDM_SETTINGS_CANCEL:
      EndDialog(hWnd, IDM_SETTINGS_CANCEL);
      return TRUE;
    default:
      return TRUE;
    }
  default:
    return FALSE;
  }
  return FALSE;
}

static main_clear()
{
  int i,j;
  for(i=0;i<nxny;i++)
  for(j=0;j<nxny;j++)
  displayboard[i][j] = 0;
}

static void window_size(int nx, int ny, int *xsize, int *ysize)
{
  *xsize = nx*ny*CELLSIZE+(nx-1)*THICKWIDTH+(nx*ny-nx)*THINWIDTH;
  *ysize = nx*ny*CELLSIZE+(ny-1)*THICKWIDTH+(nx*ny-ny)*THINWIDTH;
}

main_update()
{
  HDC hdc;
  HPEN hpen, hpenold;
  HBRUSH hbr, hbrold;
  char txt[16];
  int row,col,xsize,ysize,x,xline,y,width;

  hdc = GetDC(mainhwnd);

  window_size(nx, ny, &xsize, &ysize);

  SetBkColor(hdc, RGB(255,255,255));

  y=0;
  xline=0;
  for(row=0;row<nxny;row++)
  {
    if(row>0)
    {
      width = (row%ny==0)?THICKWIDTH:THINWIDTH;
      hpen = CreatePen(PS_SOLID, width, RGB(0,0,0));
      hpenold = SelectObject(hdc, hpen);
      MoveToEx(hdc, xline, 0, NULL);
      LineTo(hdc, xline, ysize);
      SelectObject(hdc, hpenold);
      DeleteObject(hpen);
      width = (row%nx==0)?THICKWIDTH:THINWIDTH;
      hpen = CreatePen(PS_SOLID, width, RGB(0,0,0));
      hpenold = SelectObject(hdc, hpen);
      MoveToEx(hdc, 0, y, NULL);
      LineTo(hdc, xsize, y);
      SelectObject(hdc, hpenold);
      DeleteObject(hpen);
      y+=(row%nx==0)?THICKWIDTH:THINWIDTH;
      xline+=(row%ny==0)?THICKWIDTH:THINWIDTH;
    }
    x=0;
    for(col=0;col<nxny;col++)
    {
      txt[1]=0;
      if(displayboard[row][col])
      {
        txt[0]=symbollist[abs(displayboard[row][col])];
      }
      else
      {
	txt[0]=' ';
      }
      oldfont=SelectObject(hdc, font);
      SetTextColor(hdc, displayboard[row][col]<0?RGB(255,0,0):RGB(0,0,0));
      TextOut(hdc, x+(CELLSIZE-font_width)/2, y, txt, 1);
      SelectObject(hdc, oldfont);
      if(col>0) x+=(col%ny==0)?THICKWIDTH:THINWIDTH;
      x+=CELLSIZE;
    }
    y+=CELLSIZE;
    xline+=CELLSIZE;
  }
  ReleaseDC(mainhwnd, hdc);
}

static void wincopy(HWND hWnd)
{
  LPTSTR textboard=" . . . | . . . | . . . \r\n\
 . . . | . . . | . . . \r\n\
 . . . | . . . | . . . \r\n\
-------+-------+------ \r\n\
 . . . | . . . | . . . \r\n\
 . . . | . . . | . . . \r\n\
 . . . | . . . | . . . \r\n\
-------+-------+------ \r\n\
 . . . | . . . | . . . \r\n\
 . . . | . . . | . . . \r\n\
 . . . | . . . | . . . \r\n";
  int i,r,c;
  LPTSTR lptstrCopy;
  HGLOBAL hglbCopy;
  if(OpenClipboard(hWnd))
  {
    EmptyClipboard();
    hglbCopy = GlobalAlloc(GMEM_MOVEABLE, strlen(textboard)+1);
    if(hglbCopy != NULL)
    {
      lptstrCopy = GlobalLock(hglbCopy);
      r=0;
      c=0;
      i=0;
      for(i=0; i<=strlen(textboard); i++)
      {
        if(textboard[i]=='.')
	{
	  if(displayboard[r][c]>0)
	  {
	    lptstrCopy[i] = '0'+displayboard[r][c];
	  }
	  else
	  {
	    lptstrCopy[i] = '.';
	  }
	  c++;
	  if(c==nxny)
	  {
	    c=0;
	    r++;
	  }
	}
	else
	{
	  lptstrCopy[i] = textboard[i];
	}
      }
      GlobalUnlock(hglbCopy);
      SetClipboardData(CF_TEXT, hglbCopy);
    }
    CloseClipboard();
  }
}

static void winpaste(HWND hWnd)
{
  UINT format;
  HGLOBAL hglb;
  LPTSTR lptstr;
  int i,r,c,formatok;
  TCHAR ch;
  int newboard[MAXNXNY][MAXNXNY];
  if(IsClipboardFormatAvailable(CF_TEXT))
  {
    if(OpenClipboard(hWnd))
    {
      hglb = GetClipboardData(CF_TEXT);
      if(hglb != NULL)
      {
        formatok=0;
        lptstr = GlobalLock(hglb);
	if(lptstr != NULL)
	{
	  i=0;
	  formatok=1;
	  r=0;
	  c=0;
	  while(ch=lptstr[i++])
	  {
	    if(ch=='.' || (ch>='0' && ch<='9') || (ch>='A' && ch<='Z'))
	    {
	      if(ch=='.') ch='0';
	      if(r>=nxny)
	      {
	        formatok=0;
	      }
	      else
	      {
	        if(ch>='A' && ch<='Z')
		  newboard[r][c] = ch-'A'+10;
		else
		  newboard[r][c] = ch-'0';
		c++;
		if(c==nxny)
		{
		  c=0;
		  r++;
		}
	      }
	    }
	  }
	  if(r!=nxny || c!=0) formatok=0;
	}
	GlobalUnlock(hglb);
	for(r=0;r<nxny;r++)
	for(c=0;c<nxny;c++)
	displayboard[r][c] = formatok?newboard[r][c]:0;
	main_update();
      }
      CloseClipboard();
    }
  }
}

void informuser(char *txt)
{
  MessageBox(mainhwnd, txt, "Information", MB_OK|MB_ICONWARNING|MB_APPLMODAL);
}

LRESULT CALLBACK mainWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  HDC hdc;
  SIZE tsize;
  PAINTSTRUCT ps;
  OPENFILENAME openfile;
  WINDOWINFO winfo;
  char filename[200];
  char windowtitle[80];
  int status;
  int i,j,total,ok;
  int old_width, old_height, new_width, new_height;
  FILE *statfile;
  double speed;

  switch(message)
  {
  case WM_PAINT:
    hdc = BeginPaint(hWnd, &ps);

    font = CreateFont(default_textsize,0,0,0,400,0,0,0,ANSI_CHARSET,OUT_TT_PRECIS,CLIP_DEFAULT_PRECIS,
		      DEFAULT_QUALITY,FIXED_PITCH|FF_MODERN,"Courier New");
    oldfont=SelectObject(hdc, font);
    GetTextExtentPoint32(hdc, "W", 1, &tsize);
    SelectObject(hdc, oldfont);
    font_width = tsize.cx;
    font_height = tsize.cy;
    EndPaint(hWnd, &ps);
    main_update();
    break;
  case WM_COMMAND:
    switch(LOWORD(wParam))
    {
    case IDM_MAIN_FILE_SETTINGS:
      if(DialogBox(mainhInstance, "SETTINGSDIALOG", mainhwnd, SettingsDlgProc)==IDM_SETTINGS_OK)
      {
        nfilled = newnfilled;
	symmetrical = newsymmetrical;
	level = newlevel;

        window_size(nx, ny, &old_width, &old_height);
        window_size(newnx, newny, &new_width, &new_height);
	nx=newnx;
	ny=newny;
	nxny=nx*ny;

	init_bitmaps(nxny);

	winfo.cbSize = sizeof(winfo);
	GetWindowInfo(mainhwnd, &winfo);

	SetWindowPos(mainhwnd, HWND_TOP, 0, 0, winfo.rcWindow.right - winfo.rcWindow.left+1 - old_width + new_width, 
	                                       winfo.rcWindow.bottom - winfo.rcWindow.top+1 - old_height + new_height, SWP_NOMOVE|SWP_NOZORDER);

	main_clear();
        bestndigits=0;
	SetWindowText(mainhwnd, "Sudoku Generator");
	main_update();
      }
      break;
    case IDM_MAIN_FILE_PRINT:
      if(DialogBox(mainhInstance, "PRINTERDIALOG", mainhwnd, PrinterDlgProc)==IDM_PRINTER_PRINT)
      {
        printer_print(mainhwnd);
      }
      break;
    case IDM_MAIN_EDIT_COPY:
      wincopy(hWnd);
      break;
    case IDM_MAIN_EDIT_PASTE:
      if(!running)
      {
        if(solved)
	{
	  unsolve_sudoku();
	  main_update();
	  setsolvemenutext("Solve");
	}
	winpaste(hWnd);
      }
      break;
    case IDM_MAIN_FILE_CREATE:
      if(running)
      {
	setmenutext(IDM_MAIN_FILE_CREATE, "Create");
	running=0;
	time(&stop_time);
	statfile = fopen("sudoku.stat", "w");
	if(statfile != NULL)
	{
	  total=0;
	  fprintf(statfile, "Filled\tVEasy\tEasy\tMedium\tHard\tDifficult\n");
	  for(i=0; i<81; i++)
	  {
	    ok=0;
	    for(j=0;j<5;j++)
	    {
	      if(statistics[i][j]>0) ok=1;
	    }
	    if(ok)
	    {
	      fprintf(statfile, "%2d", i);
	      for(j=0;j<5;j++)
	      {
		fprintf(statfile, "\t%d", statistics[i][j]);
		total += statistics[i][j];
	      }
	      fprintf(statfile, "\n");
	    }
	  }
	  speed=stop_time-start_time;
	  fprintf(statfile, "%d Sudokus created in %d seconds.\n", total, (int) (stop_time-start_time));
	  fprintf(statfile, "Speed: %.0f sudokus created per second.\n", ((double)total)/speed);
	  fclose(statfile);
	}
	sprintf(windowtitle, "Sudoku Generator. Stopped at iteration %d, best: %d", ntest, bestndigits);
	SetWindowText(mainhwnd, windowtitle);
      }
      else
      { 
        for(i=0;i<81;i++)
	for(j=0;j<10;j++) statistics[i][j]=0;
	time(&start_time);
	setmenutext(IDM_MAIN_FILE_CREATE, "Stop!");
	generate_sudoku();
	setmenutext(IDM_MAIN_FILE_SOLVE, "Solve");
      }
      break;
    case IDM_MAIN_FILE_SOLVE:
      if(!running)
      {
        if(solved)
	{
	  unsolve_sudoku();
	  main_update();
	  setsolvemenutext("Solve");
	}
	else
	{
	  solution_unique=solve_sudoku(1);
	  if(solution_unique==0)
	  {
	    informuser("This sudoku does not have a solution.");
	  }
	  else if(solution_unique>1)
	  {
	    informuser("This sudoku has more than one solution.");
	  }
	  main_update();
	  if(solved) setsolvemenutext("Unsolve");
	}
      }
      break;
    case IDM_MAIN_FILE_QUIT:
      exit(0);
      break;
    }
    break;
  case WM_DESTROY:
    DeleteObject(font);
    ExitProcess(0);
    break;
  default:
    return(DefWindowProc(hWnd, message, wParam, lParam));
  }
  return 0;
}

static printer_setup1st()
{
  PAGESETUPDLG pdlg;
  LPDEVNAMES lpdevnames;
  /* Set up default printer */
  memset(&pdlg, 0, sizeof(pdlg));
  pdlg.lStructSize = sizeof(pdlg);
  pdlg.hwndOwner = mainhwnd;
  pdlg.Flags = PSD_RETURNDEFAULT|PSD_MARGINS|PSD_INTHOUSANDTHSOFINCHES;
  pdlg.rtMargin.top = 500;
  pdlg.rtMargin.bottom = 500;
  pdlg.rtMargin.left = 500;
  pdlg.rtMargin.right = 500;
  PageSetupDlg(&pdlg);
  printer_lastMargin.top = pdlg.rtMargin.top;
  printer_lastMargin.bottom = pdlg.rtMargin.bottom;
  printer_lastMargin.left = pdlg.rtMargin.left;
  printer_lastMargin.right = pdlg.rtMargin.right;
  printer_marginunits = pdlg.Flags&(PSD_INTHOUSANDTHSOFINCHES|PSD_INHUNDREDTHSOFMILLIMETERS);
  printer_hDevMode = pdlg.hDevMode;
  printer_hDevNames = pdlg.hDevNames;
}


static void sudokumain()
{
  MSG Msg;
  BOOL status;

  srandom(time(NULL));

  printer_setup1st();

AGAIN:
  
  status=PeekMessage(&Msg, NULL, 0, 0, PM_REMOVE);

AGAIN2:

  if(status)
  {
    TranslateMessage(&Msg);
    DispatchMessage(&Msg);
    goto AGAIN;
  }
  if(running)
  {
    sudoku_run();
    goto AGAIN;
  }
  status=GetMessage(&Msg, NULL, 0, 0);
  if(status == -1) exit(-1);
  if(status) goto AGAIN2;

  exit(0);
}

static BOOL mainInitApplication(HINSTANCE hInstance)
{
  WNDCLASS wc;
  BOOL rc;

  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = mainWndProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInstance;
  wc.hIcon = NULL;
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
  wc.lpszMenuName = "MAINMENU";
  wc.lpszClassName = "sudokumain";

  rc=RegisterClass(&wc);
  return rc;
}

static BOOL mainInitInstance(HINSTANCE hinstance, INT ncmdshow)
{
  RECT warea;
  int xsize, ysize;

  main_clear();

  InitCommonControls();

  window_size(nx, ny, &xsize, &ysize);
  warea.left = 0;
  warea.right = xsize-1;
  warea.top = 0;
  warea.bottom = ysize-1;
  AdjustWindowRect(&warea, WS_OVERLAPPEDWINDOW, TRUE);

  mainhwnd = CreateWindow("sudokumain", "Sudoku Generator", WS_OVERLAPPEDWINDOW,
         25, 15, warea.right-warea.left+1, warea.bottom-warea.top, /* subtract one from height */
	 (HWND)NULL,(HMENU)NULL, hinstance, (LPSTR)NULL);

  if(mainhwnd == NULL) return 0;

  ShowWindow(mainhwnd, ncmdshow);
  UpdateWindow(mainhwnd);

  return 1;
}

int WINAPI WinMain(HINSTANCE hInst,HINSTANCE hPreInst, LPSTR lpszCmdLine,int nCmdShow)
{
  mainhInstance = hInst;

  mainInitApplication(mainhInstance);
  mainInitInstance(mainhInstance, nCmdShow);

  init_bitmaps(nxny);

  sudokumain();
}

